@@ -1,5 +1,5 @@
 /*
- * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
+ * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
  * @license AGPL-3.0
  *
  * This code is free software: you can redistribute it and/or modify
@@ -29,8 +29,11 @@ import java.util.HashMap;
 import java.util.LinkedHashMap;
 import java.util.List;
 import java.util.Map;
+import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicReference;
 import java.util.concurrent.locks.LockSupport;
 import javax.mail.FetchProfile;
 import javax.mail.Flags;
@@ -45,6 +48,7 @@ import org.slf4j.Logger;
 import com.openexchange.context.ContextService;
 import com.openexchange.exception.OXException;
 import com.openexchange.groupware.contexts.Context;
+import com.openexchange.groupware.contexts.UpdateBehavior;
 import com.openexchange.imap.IMAPAccess;
 import com.openexchange.imap.IMAPCapabilities;
 import com.openexchange.imap.IMAPProvider;
@@ -62,7 +66,7 @@ import com.openexchange.mail.mime.MimeTypes;
 import com.openexchange.mail.mime.utils.MimeMessageUtility;
 import com.openexchange.mail.service.MailService;
 import com.openexchange.mail.utils.MailFolderUtility;
-import com.openexchange.mailaccount.MailAccount;
+import com.openexchange.mailaccount.Account;
 import com.openexchange.mailaccount.MailAccountExceptionCodes;
 import com.openexchange.pns.DefaultPushNotification;
 import com.openexchange.pns.KnownTopic;
@@ -79,7 +83,10 @@ import com.openexchange.push.imapidle.locking.ImapIdleClusterLock;
 import com.openexchange.push.imapidle.locking.SessionInfo;
 import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
+import com.openexchange.session.Origin;
 import com.openexchange.session.Session;
+import com.openexchange.sessiond.SessionExceptionCodes;
+import com.openexchange.sessiond.SessiondService;
 import com.openexchange.threadpool.ThreadRenamer;
 import com.openexchange.timer.ScheduledTimerTask;
 import com.openexchange.timer.TimerService;
@@ -158,8 +165,11 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
     private final ServiceLookup services;
     private final ImapIdleControl control;
-    private final Session session;
-    private ScheduledTimerTask timerTask;
+    private final Session optSession;
+    private final String sessionId;
+    private final int userId;
+    private final int contextId;
+    private final AtomicReference<ScheduledTimerTask> timerTask;
     private final int accountId;
     private final String fullName;
     private final long delayNanos;
@@ -167,8 +177,9 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
     private final AtomicBoolean canceled;
     private final boolean permanent;
     private final boolean supportsPermanentListeners;
-    private volatile IMAPFolder imapFolderInUse;
-    private volatile Map<String, Object> additionalProps;
+    private final AtomicReference<IMAPFolder> imapFolderInUse;
+    private final AtomicReference<Map<String, Object>> additionalProps;
+    private final ConcurrentMap<Thread, Session> notifySessions;
     private volatile boolean interrupted;
     private volatile boolean idling;
 
@@ -178,22 +189,28 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
     public ImapIdlePushListener(String fullName, int accountId, PushMode pushMode, long delay, Session session, boolean permanent, boolean supportsPermanentListeners, ImapIdleControl control, ServiceLookup services) {
         super();
         canceled = new AtomicBoolean();
+        imapFolderInUse = new AtomicReference<>();
+        additionalProps = new AtomicReference<>();
+        notifySessions = new ConcurrentHashMap<>();
+        timerTask = new AtomicReference<>();
         this.permanent = permanent;
         this.supportsPermanentListeners = supportsPermanentListeners;
         this.fullName = fullName;
         this.accountId = accountId;
-        this.session = session;
+        this.optSession = session.getOrigin() == Origin.SYNTHETIC ? session : null;
+        this.sessionId = session.getSessionID();
+        this.userId = session.getUserId();
+        this.contextId = session.getContextId();
         this.delayNanos = TimeUnit.MILLISECONDS.toNanos(delay <= 0 ? 5000L : delay);
         this.services = services;
         this.control = control;
         this.pushMode = pushMode;
-        additionalProps = null;
     }
 
-    private boolean isUserValid() {
+    private boolean isUserValid(Session session) {
         try {
             ContextService contextService = services.getService(ContextService.class);
-            Context context = contextService.loadContext(session.getContextId());
+            Context context = contextService.loadContext(session.getContextId(), UpdateBehavior.DENY_UPDATE);
             if (!context.isEnabled()) {
                 return false;
             }
@@ -212,7 +229,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
      * @return The context identifier
      */
     public int getContextId() {
-        return session.getContextId();
+        return contextId;
     }
 
     /**
@@ -221,7 +238,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
      * @return The user identifier
      */
     public int getUserId() {
-        return session.getUserId();
+        return userId;
     }
 
     /**
@@ -251,7 +268,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
     @Override
     public void notifyNewMail() throws OXException {
-        Map<String, Object> props = this.additionalProps;
+        Map<String, Object> props = this.additionalProps.get();
 
         PushNotificationService pushNotificationService = services.getOptionalService(PushNotificationService.class);
         if (null != pushNotificationService) {
@@ -263,6 +280,10 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             }
         }
 
+        Session session = notifySessions.get(Thread.currentThread());
+        if (session == null) {
+            throw OXException.general("No such session for thread: " + Thread.currentThread().getName());
+        }
         if (null == props) {
             props = new LinkedHashMap<>(2);
         }
@@ -271,9 +292,6 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
     }
 
     private Collection<PushNotification> createNotification(Map<String, Object> props) {
-        int userId = session.getUserId();
-        int contextId = session.getContextId();
-
         if (null == props) {
             Map<String, Object> messageData = new LinkedHashMap<>(2);
             messageData.put(PushNotificationField.FOLDER.getId(), MailFolderUtility.prepareFullname(accountId, fullName));
@@ -352,7 +370,26 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             return;
         }
 
-        if (!isUserValid()) {
+        Session session = optSession;
+        if (session == null) {
+            // Get required sessiond service
+            SessiondService sessiondService = services.getOptionalService(SessiondService.class);
+            if (sessiondService == null) {
+                cancel(false);
+                return;
+            }
+
+            // Ensure session still valid
+            session = sessiondService.getSession(sessionId);
+            if (session == null) {
+                // No such session
+                cancel(true);
+                return;
+            }
+        }
+
+        // Check user validity
+        if (!isUserValid(session)) {
             cancel(false);
             return;
         }
@@ -384,9 +421,9 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
                 IMAPStore imapStore = IMAPAccess.getIMAPFolderStorageFrom(mailAccess1).getImapStore();
                 final IMAPFolder imapFolder = (IMAPFolder) imapStore.getFolder(fullName);
-                this.imapFolderInUse = imapFolder;
+                this.imapFolderInUse.set(imapFolder);
                 Map<String, Object> props = new HashMap<String, Object>(3);
-                this.additionalProps = props;
+                this.additionalProps.set(props);
                 long uidNext = -1;
                 try {
                     // The next expected UID
@@ -411,7 +448,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                     }
 
                     // Refresh lock prior to entering IMAP-IDLE
-                    ImapIdlePushManagerService.getInstance().refreshLock(new SessionInfo(session, permanent, permanent ? false : isTransient(session, services)));
+                    ImapIdlePushManagerService.getInstance().refreshLock(new SessionInfo(sessionId, userId, contextId, permanent, permanent ? false : isTransient(session, services)));
 
                     // Are there already new messages?
                     {
@@ -422,7 +459,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             if (recentMessages != null && recentMessages.length > 0) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
                                 setEventPropertiesByRecent(recentMessages, props, imapFolder);
-                                notifyNewMail();
+                                doNotifyNewMail(session);
                                 LOGGER.debug("Performed IMAP-IDLE run having new messages for user {} in context {}. ", sUserId, sContextId);
                                 return;
                             }
@@ -431,14 +468,14 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
                     // Do the IMAP IDLE connect
                     {
-                        long st = System.nanoTime();
+                        long st = System.nanoTime(); // NOSONARLINT
 
                         mailAccess.setWaiting(true);
                         try {
                             if (false == doImapIdle(imapFolder)) {
                                 // Timeout elapsed
                                 error = false;
-                                return;
+                                return; // NOSONARLINT Continue with next timer run
                             }
                         } finally {
                             mailAccess.setWaiting(false);
@@ -464,7 +501,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             if (newMessageCount > 0) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
                                 setEventProperties(uidNext, totalCount, props, imapFolder);
-                                notifyNewMail();
+                                doNotifyNewMail(session);
                                 notified = true;
                             }
                         }
@@ -478,7 +515,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             if (newMessageCount > 0) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
                                 setEventProperties(uidNext, totalCount, props, imapFolder);
-                                notifyNewMail();
+                                doNotifyNewMail(session);
                                 notified = true;
                                 break;
                             }
@@ -490,7 +527,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             if (imapFolder.getDeletedMessageCount() != deletedCount) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to differing message counts. Current deleted count {} vs. old deleted count {}", sUserId, sContextId, Integer.toString(newDeletedCount), Integer.toString(deletedCount));
                                 setEventProperties(uidNext, totalCount, props, imapFolder);
-                                notifyNewMail();
+                                doNotifyNewMail(session);
                                 notified = true;
                                 break;
                             }
@@ -502,7 +539,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                             if (newTotalCount != totalCount) {
                                 LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to differing message counts. Current total count {} vs. old total count {}", sUserId, sContextId, Integer.toString(newTotalCount), Integer.toString(totalCount));
                                 setEventProperties(uidNext, totalCount, props, imapFolder);
-                                notifyNewMail();
+                                doNotifyNewMail(session);
                                 notified = true;
                                 break;
                             }
@@ -515,13 +552,13 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                         LOGGER.debug("Performed IMAP-IDLE run with no result for user {} in context {}. ", sUserId, sContextId);
                     }
                 } finally {
-                    this.imapFolderInUse = null;
+                    this.imapFolderInUse.set(null);
                     try {
                         imapFolder.close(false);
                     } catch (Exception e) {
                         // Ignore
                     }
-                    this.additionalProps = null;
+                    this.additionalProps.set(null);
                 }
             } catch (OXException e) {
                 launderOXException(e);
@@ -557,6 +594,21 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         }
     }
 
+    /**
+     * Invokes {@link #notifyNewMail()} while remembering given session temporarily in appropriate map.
+     *
+     * @param session The session to remember for the invocation
+     * @throws OXException If notification fails
+     */
+    private void doNotifyNewMail(Session session) throws OXException {
+        notifySessions.put(Thread.currentThread(), session);
+        try {
+            notifyNewMail();
+        } finally {
+            notifySessions.remove(Thread.currentThread());
+        }
+    }
+
     /**
      * Actually enters the IMAP-IDLE to IMAP server with respect to frequent cluster lock <i><tt>touch</tt></i>ing.
      * <p>
@@ -625,7 +677,22 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             if (null == mailService) {
                 throw ServiceExceptionCode.absentService(MailService.class);
             }
-            access = mailService.getMailAccess(session, MailAccount.DEFAULT_ID);
+
+            Session session = optSession;
+            if (session == null) {
+                SessiondService sessiondService = services.getOptionalService(SessiondService.class);
+                if (sessiondService == null) {
+                    throw ServiceExceptionCode.absentService(SessiondService.class);
+                }
+
+                session = sessiondService.getSession(sessionId);
+                if (session == null) {
+                    // No such session
+                    throw SessionExceptionCodes.SESSION_EXPIRED.create(sessionId);
+                }
+            }
+
+            access = mailService.getMailAccess(session, Account.DEFAULT_ID);
             /*
              * Check protocol
              */
@@ -653,12 +720,13 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         }
 
         long delay = TimeUnit.NANOSECONDS.toMillis(delayNanos);
-        timerTask = timerService.scheduleWithFixedDelay(this, delay, delay);
+        timerTask.set(timerService.scheduleWithFixedDelay(this, delay, delay));
     }
 
     /**
      * Cancels this IMAP IDLE listener.
      *
+     * @param tryToReconnect <code>true</code> to try reconnecting IMAP IFLE listener; otherwise <code>false</code>
      * @return <code>null</code> if reconnected; otherwise a clean-up task if terminated
      */
     public synchronized Runnable cancel(boolean tryToReconnect) {
@@ -668,9 +736,8 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             canceled.set(true);
 
             // Close IMAP resources, too
-            final IMAPFolder imapFolderInUse = this.imapFolderInUse;
+            IMAPFolder imapFolderInUse = this.imapFolderInUse.getAndSet(null);
             if (null != imapFolderInUse) {
-                this.imapFolderInUse = null;
                 try {
                     imapFolderInUse.close(false);
                 } catch (Exception e) {
@@ -679,15 +746,18 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             }
 
             // Cancel timer task
-            ScheduledTimerTask timerTask = this.timerTask;
+            ScheduledTimerTask timerTask = this.timerTask.getAndSet(null);
             if (null != timerTask) {
-                this.timerTask = null;
                 timerTask.cancel();
+                TimerService timerService = services.getOptionalService(TimerService.class);
+                if (null != timerService) {
+                    timerService.purge();
+                }
             }
         } finally {
             ImapIdlePushManagerService instance = ImapIdlePushManagerService.getInstance();
             if (null != instance) {
-                ImapIdlePushListener anotherListener = tryToReconnect ? instance.injectAnotherListenerFor(session) : null;
+                ImapIdlePushListener anotherListener = tryToReconnect ? instance.injectAnotherListenerFor(sessionId, userId, contextId) : null;
                 if (null == anotherListener) {
                     // No other listener available
                     // Give up lock and return
@@ -696,7 +766,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                     try {
                         anotherListener.start();
                     } catch (Exception e) {
-                        LOGGER.warn("Failed to start new listener for user {} in context {}.", Integer.valueOf(session.getUserId()), Integer.valueOf(session.getContextId()), e);
+                        LOGGER.warn("Failed to start new listener for user {} in context {}.", Integer.valueOf(userId), Integer.valueOf(contextId), e);
                         // Give up lock and return
                         cleanUpTask = createCleanUpTask(instance);
                     }
@@ -706,18 +776,27 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         return cleanUpTask;
     }
 
-    private Runnable createCleanUpTask(final ImapIdlePushManagerService instance) {
-        final Session session = this.session;
-        final boolean permanent = this.permanent;
-        final ServiceLookup services = this.services;
-        return new Runnable() {
-            @Override
-            public void run() {
-                try {
-                    instance.releaseLock(new SessionInfo(session, permanent, permanent ? false : isTransient(session, services)));
-                } catch (Exception e) {
-                    LOGGER.warn("Failed to release lock for user {} in context {}.", Integer.valueOf(session.getUserId()), Integer.valueOf(session.getContextId()), e);
-                }
+    private Runnable createCleanUpTask(ImapIdlePushManagerService instance) {
+        String sessionId = this.sessionId;
+
+        Session session = optSession;
+        if (session == null) {
+            // Get required sessiond service
+            SessiondService sessiondService = services.getOptionalService(SessiondService.class);
+            if (sessiondService != null) {
+                session = sessiondService.getSession(sessionId);
+            }
+        }
+        Session ses = session;
+
+        int userId = this.userId;
+        int contextId = this.contextId;
+        boolean permanent = this.permanent;
+        return () -> {
+            try {
+                instance.releaseLock(new SessionInfo(sessionId, userId, contextId, permanent, !permanent && (ses != null && isTransient(ses, services))));
+            } catch (Exception e) {
+                LOGGER.warn("Failed to release lock for user {} in context {}.", Integer.valueOf(userId), Integer.valueOf(contextId), e);
             }
         };
     }
@@ -730,7 +809,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             /*
              * Missing mail account; drop listener
              */
-            LOGGER.debug("Missing (default) mail account for user {} in context {}. Stopping obsolete IMAP-IDLE listener.", Integer.valueOf(session.getUserId()), Integer.valueOf(session.getContextId()));
+            LOGGER.debug("Missing (default) mail account for user {} in context {}. Stopping obsolete IMAP-IDLE listener.", Integer.valueOf(userId), Integer.valueOf(contextId));
             throw e;
         }
         if ("DBP".equals(e.getPrefix())) {
@@ -919,6 +998,7 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             LOGGER.debug("Failed to get \"Bcc\" address header", x);
             mailMessage.addBcc(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_BCC, im));
         }
+        mailMessage.addSender(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_SENDER, im));
 
         final Flags msgFlags = im.getFlags();
         int flags = 0;
