@@ -1,5 +1,5 @@
 /*
- * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
+ * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
  * @license AGPL-3.0
  *
  * This code is free software: you can redistribute it and/or modify
@@ -43,6 +43,7 @@ import com.openexchange.exception.OXException;
 import com.openexchange.hazelcast.Hazelcasts;
 import com.openexchange.java.Streams;
 import com.openexchange.lock.AccessControl;
+import com.openexchange.lock.AccessControls;
 import com.openexchange.lock.LockService;
 import com.openexchange.lock.ReentrantLockAccessControl;
 import com.openexchange.push.PushListener;
@@ -62,6 +63,7 @@ import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
 import com.openexchange.session.ObfuscatorService;
 import com.openexchange.session.Session;
+import com.openexchange.session.UserAndContext;
 import com.openexchange.sessiond.SessionMatcher;
 import com.openexchange.sessiond.SessiondService;
 import com.openexchange.sessiond.SessiondServiceExtended;
@@ -119,7 +121,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
     private final String name;
     private final ServiceLookup services;
     private final ImapIdleControl control;
-    private final ConcurrentMap<SimpleKey, ImapIdlePushListener> listeners;
+    private final ConcurrentMap<UserAndContext, ImapIdlePushListener> listeners;
     private final String fullName;
     private final int accountId;
     private final PushMode pushMode;
@@ -141,7 +143,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
         this.clusterLock = clusterLock;
         this.services = services;
         this.control = new ImapIdleControl();
-        listeners = new ConcurrentHashMap<SimpleKey, ImapIdlePushListener>(512, 0.9f, 1);
+        listeners = new ConcurrentHashMap<UserAndContext, ImapIdlePushListener>(512, 0.9F, 1);
 
         if (checkPeriodic) {
             // Initialize timer task to check for expired IMAP-IDLE push listeners for every 30 seconds
@@ -257,9 +259,9 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
     @Override
     public List<PushUserInfo> getAvailablePushUsers() throws OXException {
         List<PushUserInfo> l = new LinkedList<PushUserInfo>();
-        for (Map.Entry<SimpleKey, ImapIdlePushListener> entry : listeners.entrySet()) {
-            SimpleKey key = entry.getKey();
-            l.add(new PushUserInfo(new PushUser(key.userId, key.contextId), entry.getValue().isPermanent()));
+        for (Map.Entry<UserAndContext, ImapIdlePushListener> entry : listeners.entrySet()) {
+            UserAndContext key = entry.getKey();
+            l.add(new PushUserInfo(new PushUser(key.getUserId(), key.getContextId()), entry.getValue().isPermanent()));
         }
         return l;
     }
@@ -290,7 +292,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
             return null;
         }
 
-        SessionInfo sessionInfo = new SessionInfo(session, true, false);
+        SessionInfo sessionInfo = new SessionInfo(session.getSessionID(), session.getUserId(), session.getContextId(), true, false);
         AcquisitionResult acquisitionResult = clusterLock.acquireLock(sessionInfo);
         if (AcquisitionResult.NOT_ACQUIRED == acquisitionResult) {
             LOGGER.info("Could not acquire lock to start permanent IMAP-IDLE listener for user {} in context {} with session {} as there is already an associated listener", I(userId), I(contextId), session.getSessionID());
@@ -316,7 +318,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
                 try {
                     lock.acquireGrant();
                     ImapIdlePushListener listener = new ImapIdlePushListener(fullName, accountId, pushMode, delay, session, true, supportsPermanentListeners(), control, services);
-                    ImapIdlePushListener current = listeners.putIfAbsent(SimpleKey.valueOf(userId, contextId), listener);
+                    ImapIdlePushListener current = listeners.putIfAbsent(UserAndContext.newInstance(userId, contextId), listener);
                     if (null == current) {
                         listener.start();
                         unlock = false;
@@ -325,7 +327,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
                     } else if (!current.isPermanent()) {
                         // Cancel current & replace
                         current.cancel(false);
-                        listeners.put(SimpleKey.valueOf(userId, contextId), listener);
+                        listeners.put(UserAndContext.newInstance(userId, contextId), listener);
                         listener.start();
                         unlock = false;
                         LOGGER.info("Started permanent IMAP-IDLE listener for user {} in context {}", I(userId), I(contextId));
@@ -391,7 +393,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
             return null;
         }
 
-        SessionInfo sessionInfo = new SessionInfo(session, false, isTransient(session, services));
+        SessionInfo sessionInfo = new SessionInfo(session.getSessionID(), session.getUserId(), session.getContextId(), false, isTransient(session, services));
         AcquisitionResult acquisitionResult = clusterLock.acquireLock(sessionInfo);
         if (AcquisitionResult.NOT_ACQUIRED == acquisitionResult) {
             LOGGER.info("Could not acquire lock to start IMAP-IDLE listener for user {} in context {} with session {} as there is already an associated listener", I(userId), I(contextId), session.getSessionID());
@@ -417,7 +419,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
                 try {
                     lock.acquireGrant();
                     ImapIdlePushListener listener = new ImapIdlePushListener(fullName, accountId, pushMode, delay, session, false, supportsPermanentListeners(), control, services);
-                    if (null == listeners.putIfAbsent(SimpleKey.valueOf(userId, contextId), listener)) {
+                    if (null == listeners.putIfAbsent(UserAndContext.newInstance(userId, contextId), listener)) {
                         listener.start();
                         unlock = false;
                         LOGGER.info("Started IMAP-IDLE listener for user {} in context {} with session {} ({})", I(userId), I(contextId), session.getSessionID(), session.getClient());
@@ -477,11 +479,13 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
      * @return The stop result
      */
     public StopResult stopListener(boolean tryToReconnect, boolean stopIfPermanent, int userId, int contextId) {
-        AccessControl lock = getlockFor(userId, contextId);
         Runnable cleanUpTask = null;
+        boolean acquired = false;
+        AccessControl lock = getlockFor(userId, contextId);
         try {
             lock.acquireGrant();
-            SimpleKey key = SimpleKey.valueOf(userId, contextId);
+            acquired = true;
+            UserAndContext key = UserAndContext.newInstance(userId, contextId);
             ImapIdlePushListener listener = listeners.get(key);
             if (null != listener) {
                 if (!stopIfPermanent && listener.isPermanent()) {
@@ -505,7 +509,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
             Thread.currentThread().interrupt();
             throw new IllegalStateException(e);
         } finally {
-            Streams.close(lock);
+            AccessControls.release(lock, acquired);
             if (null != cleanUpTask) {
                 cleanUpTask.run();
             }
@@ -535,14 +539,13 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
     /**
      * Tries to look-up another valid session and injects a new listener for it (discarding the existing one bound to given <code>oldSession</code>)
      *
-     * @param oldSession The expired/outdated session
+     * @param oldSession The identifier of the expired/outdated session
+     * @param userId The user identifier
+     * @param contextId The context identifier
      * @return The new listener or <code>null</code>
      * @throws OXException If operation fails
      */
-    public ImapIdlePushListener injectAnotherListenerFor(Session oldSession) {
-        int contextId = oldSession.getContextId();
-        int userId = oldSession.getUserId();
-
+    public ImapIdlePushListener injectAnotherListenerFor(String oldSessionId, int userId, int contextId) {
         // Prefer permanent listener prior to performing look-up for another valid session
         if (supportsPermanentListeners() && hasPermanentPush(userId, contextId)) {
             try {
@@ -556,8 +559,6 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
         // Look-up sessions
         SessiondService sessiondService = services.getService(SessiondService.class);
         if (null != sessiondService) {
-            final String oldSessionId = oldSession.getSessionID();
-
             // Query local ones first
             SessionMatcher matcher = new SessionMatcher() {
 
@@ -568,9 +569,10 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
 
                 @Override
                 public boolean accepts(Session session) {
-                    return !oldSessionId.equals(session.getSessionID()) && PushUtility.allowedClient(session.getClient(), session, true);
+                    return !oldSessionId.equals(oldSessionId) && PushUtility.allowedClient(session.getClient(), session, true);
                 }
             };
+
             Session anotherActiveSession = sessiondService.findFirstMatchingSessionForUser(userId, contextId, matcher);
             if (anotherActiveSession != null) {
                 return injectAnotherListenerUsing(anotherActiveSession, false).injectedPushListener;
@@ -579,7 +581,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
             // If we're running a node-local lock, there is no need to check for sessions at other nodes
             if (ImapIdleClusterLock.Type.LOCAL != clusterLock.getType()) {
                 // Look-up remote sessions, too, if possible
-                Session session = lookUpRemoteSessionFor(oldSession);
+                Session session = lookUpRemoteSessionFor(oldSessionId, userId, contextId);
                 if (null != session) {
                     Session ses = sessiondService.getSession(session.getSessionID());
                     if (ses != null) {
@@ -592,7 +594,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
         return null;
     }
 
-    private Session lookUpRemoteSessionFor(Session oldSession) {
+    private Session lookUpRemoteSessionFor(String oldSessionId, int userId, int contextId) {
         HazelcastInstance hzInstance = services.getOptionalService(HazelcastInstance.class);
         final ObfuscatorService obfuscatorService = services.getOptionalService(ObfuscatorService.class);
         if (null == hzInstance || null == obfuscatorService) {
@@ -605,9 +607,6 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
             return null;
         }
 
-        int contextId = oldSession.getContextId();
-        int userId = oldSession.getUserId();
-        final String oldSessionId = oldSession.getSessionID();
         Hazelcasts.Filter<PortableSessionCollection, PortableSession> filter = new Hazelcasts.Filter<PortableSessionCollection, PortableSession>() {
 
             @Override
@@ -649,7 +648,7 @@ public final class ImapIdlePushManagerService implements PushManagerExtendedServ
     public InjectedImapIdlePushListener injectAnotherListenerUsing(Session newSession, boolean permanent) {
         ImapIdlePushListener listener = new ImapIdlePushListener(fullName, accountId, pushMode, delay, newSession, permanent, supportsPermanentListeners(), control, services);
         // Replace old/existing one
-        ImapIdlePushListener prev = listeners.put(SimpleKey.valueOf(newSession), listener);
+        ImapIdlePushListener prev = listeners.put(UserAndContext.newInstance(newSession), listener);
         return new InjectedImapIdlePushListener(listener, prev);
     }
 
