@@ -1,5 +1,5 @@
 /*
- * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
+ * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
  * @license AGPL-3.0
  *
  * This code is free software: you can redistribute it and/or modify
@@ -22,7 +22,6 @@
 package com.openexchange.push.imapidle.locking;
 
 import javax.annotation.concurrent.Immutable;
-import com.openexchange.session.Session;
 
 /**
  * {@link SessionInfo} - The session info for the IMAP-IDLE listener.
@@ -42,15 +41,17 @@ public class SessionInfo {
     /**
      * Initializes a new {@link SessionInfo}.
      *
-     * @param session The associated session
+     * @param sessionId The identifier of the associated session
+     * @param userId = The user identifier
+     * @param contextId = The context identifier
      * @param permanent Whether permanent or not
      * @param tranzient <code>true</code> if session is not supposed to be held in session storage; otherwise <code>false</code>
      */
-    public SessionInfo(Session session, boolean permanent, boolean tranzient) {
+    public SessionInfo(String sessionId, int userId, int contextId, boolean permanent, boolean tranzient) {
         super();
-        this.contextId = session.getContextId();
-        this.userId = session.getUserId();
-        this.sessionId = session.getSessionID();
+        this.contextId = contextId;
+        this.userId = userId;
+        this.sessionId = sessionId;
         this.permanent = permanent;
         this.tranzient = tranzient;
     }
